/*
 * "fbp.h" - header for "frame buffer pointer" object type
 *
 * written by Peter C. Jones
 * last modified 7-6-1994
 */

/************************************************************
Copyright (c) 1994  Sun Microsystems, Inc.
Copyright (c) 1993  X Consortium

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documenta-
tion files (the "Software"), to deal in the Software with-
out restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to
whom the Software is furnished to do so, subject to the 
following conditions:

The above copyright notice and this permission notice shall
be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE, AND NONINFRINGEMENT.  IN NO EVENT SHALL SUN OR THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES, OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHER-
WISE, ARISING FROM, OUT OF, OR IN CONNECTION WITH THE 
SOFTWARE OR THE USE OR DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the names of Sun
Microsystems, Inc. and the X Consortium shall not be used in
advertising or otherwise to promote the sale, use, or other
dealings in this Software without prior written authoriza-
tion from Sun Microsystems, Inc. and/or the X Consortium, as
appropriate.
*************************************************************/

#ifndef _FBP_H
#define _FBP_H

#ident "%W% Xvan 7-6-1994"

#ifdef class
#undef class
#define redefine_class
#endif

#ifdef private
#undef private
#define redefine_private
#endif

#ifdef public
#undef public
#define redefine_public
#endif

/*
 * constants
 */

// #define USE_DEBUG_OPERATORS

#define MAXFBS 9

/*
 * "FBList" class - manages list of frame buffers and their position within
 *                  a virtual frame buffer
 */
class FBList {
  friend FBPBase;
  friend FBP, FBP8;                     // these for debugging only

private:
  int width, height, depth;             // virtual frame buffer geometry
  
public: // hw/sun files need to access these internals: an interface must be 
        // made to fix this...
  int num;                              // num of actual fbs in list
  struct FBDesc {
    int x, y, width, height, right, bottom;
    unsigned int *addr;
  } fbdesc[MAXFBS];

private:  
  unsigned int wordwidth;               // width of each scan line, in words
  unsigned int tablesize;               // num of elements in lookup table
  unsigned int shift, mask;             // lookup parameters
  unsigned int **table;                 // pointer lookup table
  unsigned int *yard;                   // junkyard for unmapped areas
  
  int screenswide, screenshigh;         // dimensions of screen grid
  
public:
  FBList();
  ~FBList();

  void *operator new(size_t, void *p) { return p; }

  enum position { RIGHT, BOTTOM, LEFT, TOP };
  void addRel(int, int, int, unsigned int *, position);
  void add(int w, int h, int d, unsigned int *a) {
    addRel(w, h, d, a, RIGHT);
  }

  int getWidth() { return width; }
  int getHeight() { return height; }
  int getDepth() { return depth; }
  unsigned int *getAddr() { return table[0]; }
};

/*
 * "FBPBase" - base class for pointers into a virtual frame buffer
 */
class FBPBase {
protected:
  enum { uninitialized, fake, simple, multiple } type;
  unsigned int value;                   // virtual pointer value
  FBList *fblist;
  unsigned char **table;
  unsigned int shift, mask;

  void initfake(unsigned int *);

public:
  FBPBase();
  FBPBase(FBList *);

  unsigned int getValue() { return value; }
};

/*
 * "FBP" class - used to point to unsigned integers in a virtual frame buffer
 */
class FBP : public FBPBase {
private:
  unsigned int& debug_operator_star();
  unsigned int& debug_operator_brackets(int);

public:
  FBP() : FBPBase() {}
  FBP(FBList *fblp) : FBPBase(fblp) {}
  FBP(unsigned int *addr) : FBPBase() { initfake(addr); }
  FBP(unsigned long *addr) : FBPBase() { initfake((unsigned int *) addr); }
  FBP(FBPBase& base) : FBPBase(base) {}

  friend inline FBP operator+(const FBP& a, int x) { 
    FBP z = a;
    z.value += (x << 2); 
    return z; 
  }
  friend inline FBP operator-(const FBP& a, int x) { 
    FBP z = a;
    z.value -= (x << 2);
    return z; 
  }
  friend inline int operator-(const FBP& a, const FBP& b) {
    return (a.value - b.value) >> 2;
  }
  friend inline int operator<(const FBP& a, const FBP& b) {
    return a.value < b.value;
  }
  FBP& operator+=(int x) { value += (x << 2); return *this; }
  FBP& operator-=(int x) { value -= (x << 2); return *this; }
  FBP& operator++() { value += 4; return *this; }
  FBP operator++(int) { FBP z = *this; value += 4; return z; }
  FBP& operator--() { value -= 4; return *this; }
  FBP operator--(int) { FBP z = *this; value -= 4; return z; }
  
  unsigned int& operator*() {
#ifdef USE_DEBUG_OPERATORS
    return debug_operator_star();
#else
    return *((unsigned int *)
	     (table[value >> shift] + (value & mask)));
#endif
  }
  unsigned int& operator[](int x) {
#ifdef USE_DEBUG_OPERATORS
    return debug_operator_brackets(x);
#else
    return *((unsigned int *)
	     (table[(value + (x << 2)) >> shift] +
	      ((value + (x << 2)) & mask)));
#endif
  }
};

/*
 * "FBP32" class - used to point to 32-bit words in a virtual frame buffer
 */
typedef FBP FBP32;

/*
 * "FBP16" class - used to point to 16-bit words in a virtual frame buffer
 */
class FBP16 : public FBPBase {
private:
  unsigned short& debug_operator_star();
  unsigned short& debug_operator_brackets(int);

public:
  FBP16() : FBPBase() {}
  FBP16(FBList *fblp) : FBPBase(fblp) {}
  FBP16(unsigned short *addr) : FBPBase() { initfake((unsigned int *) addr); }
  FBP16(FBPBase& base) : FBPBase(base) {}

  friend inline FBP16 operator+(const FBP16& a, int x) { 
    FBP16 z = a;
    z.value += (x << 1); 
    return z; 
  }
  friend inline FBP16 operator-(const FBP16& a, int x) { 
    FBP16 z = a;
    z.value -= (x << 1);
    return z; 
  }
  friend inline int operator-(const FBP16& a, const FBP16& b) {
    return (a.value - b.value) >> 1;
  }
  friend inline int operator<(const FBP16& a, const FBP16& b) {
    return a.value < b.value;
  }
  FBP16& operator+=(int x) { value += (x << 1); return *this; }
  FBP16& operator-=(int x) { value -= (x << 1); return *this; }
  FBP16& operator++() { value += 2; return *this; }
  FBP16 operator++(int) { FBP16 z = *this; value += 2; return z; }
  FBP16& operator--() { value -= 2; return *this; }
  FBP16 operator--(int) { FBP16 z = *this; value -= 2; return z; }
  
  unsigned short& operator*() {
#ifdef USE_DEBUG_OPERATORS
    return debug_operator_star();
#else
    return *((unsigned short *)
	     (table[value >> shift] + (value & mask)));
#endif
  }
  unsigned short& operator[](int x) {
#ifdef USE_DEBUG_OPERATORS
    return debug_operator_brackets(x);
#else
    return *((unsigned short *)
	     (table[(value + (x << 1)) >> shift] +
	      ((value + (x << 1)) & mask)));
#endif
  }
};

/*
 * "FBP8" class - used to point to bytes in a virtual frame buffer
 */
class FBP8 : public FBPBase {
private:
  unsigned char& debug_operator_star();
  unsigned char& debug_operator_brackets(int);

public:
  FBP8() : FBPBase() {}
  FBP8(FBList *fblp) : FBPBase(fblp) {}
  FBP8(unsigned char *addr) : FBPBase() { initfake((unsigned int *) addr); }
  FBP8(FBPBase& base) : FBPBase(base) {}

  friend inline FBP8 operator+(const FBP8& a, int x) { 
    FBP8 z = a;
    z.value += x; 
    return z; 
  }
  friend inline FBP8 operator-(const FBP8& a, int x) { 
    FBP8 z = a;
    z.value -= x;
    return z; 
  }
  friend inline int operator-(const FBP8& a, const FBP8& b) {
    return a.value - b.value;
  }
  friend inline int operator<(const FBP8& a, const FBP8& b) {
    return a.value < b.value;
  }
  FBP8& operator+=(int x) { value += x; return *this; }
  FBP8& operator-=(int x) { value -= x; return *this; }
  FBP8& operator++() { ++ value; return *this; }
  FBP8 operator++(int) { FBP8 z = *this; ++ value; return z; }
  FBP8& operator--() { -- value; return *this; }
  FBP8 operator--(int) { FBP8 z = *this; -- value; return z; }
  
  unsigned char& operator*() {
#ifdef USE_DEBUG_OPERATORS
    return debug_operator_star();
#else
    return *((unsigned char *)
	     (table[value >> shift] + (value & mask)));
#endif
  }
  unsigned char& operator[](int x) {
#ifdef USE_DEBUG_OPERATORS
    return debug_operator_brackets(x);
#else
    return *((unsigned char *)
	     (table[(value + x) >> shift] +
	      ((value + x) & mask)));
#endif
  }
};

#ifdef redefine_class
#undef redefine_class
#define class _class
#endif

#ifdef redefine_private
#undef redefine_private
#define private _private
#endif

#ifdef redefine_public
#undef redefine_public
#define public _public
#endif

#endif /* _FBP_H */
